import logging
from dataclasses import dataclass
from typing import Union, Optional

from linkml_runtime import SchemaView
from linkml_runtime.linkml_model import SchemaDefinition

from schema_automator.annotators import SchemaAnnotator

logger = logging.getLogger(__name__)

SYSTEM_PROMPT = """
You are a LinkML schema assistant and your job is to suggest
values for missing elements in the schema.
Schema details: {schema_info}
"""

def enrich_using_llm(schema: Union[SchemaDefinition, str], model_name: str = None) -> SchemaDefinition:
    """
    Use the LLM model to enrich the schema with descriptions for elements that are missing them.

    Requires installation with the llm extra.

    :param schema:
    :param model_name:
    :return:
    """
    from llm import get_model, Conversation
    if not model_name:
        model_name = "gpt-4-turbo"
    model = get_model(model_name)
    sv = SchemaView(schema)
    schema = sv.schema
    schema_info =f"name: {schema.name}, description: {schema.description}"
    system_prompt = SYSTEM_PROMPT.format(schema_info=schema_info)
    logger.info(f"System: {system_prompt}")
    for elt in sv.all_elements().values():
        typ = elt.__class__.__name__
        if not elt.description:
            prompt = (f"Generate a description for the {typ} with name '{elt.name}'"
                      "The description should be a concise phrase that describes the element.")
            conversation = Conversation(model=model)
            result = conversation.prompt(prompt, system_prompt)
            elt.description = result.text()
            logger.debug(f"Updated Description: {elt.description}")
    return schema


@dataclass
class LLMAnnotator(SchemaAnnotator):
    """
    Annotates a schema using an LLM.
    """

    model_name: Optional[str] = None

    def enrich(self, schema: Union[SchemaDefinition, str]) -> SchemaDefinition:
        """
        Enrich a schema using an LLM.

        >>> from schema_automator.utils.schemautils import write_schema
        >>> annotator = LLMAnnotator(model_name="gpt-4")
        >>> schema = annotator.enrich("tests/resources/biopax3.yaml")
        >>> write_schema(schema)

        Requires installation with the llm extra:

        .. code-block:: bash

            pip install schema-automator[llm]

        Note: exercise caution on running this on large schemas with expensive
        models like gpt-4.

        :param schema:
        :return:
        """
        return enrich_using_llm(schema, model_name=self.model_name)



"""
Command Line Interface to Schema Automator
-----------------------------

"""
import logging
import os
import click


import pandas as pd

import yaml
from linkml_runtime.linkml_model import SchemaDefinition
from oaklib.selector import get_implementation_from_shorthand

from schema_automator import JsonLdAnnotator, FrictionlessImportEngine
from schema_automator.annotators import llm_annotator
from schema_automator.annotators.schema_annotator import SchemaAnnotator
from schema_automator.generalizers.csv_data_generalizer import CsvDataGeneralizer
from schema_automator.generalizers.generalizer import DEFAULT_CLASS_NAME, DEFAULT_SCHEMA_NAME
from schema_automator.generalizers.pandas_generalizer import PandasDataGeneralizer
from schema_automator.importers.dosdp_import_engine import DOSDPImportEngine
from schema_automator.generalizers.json_instance_generalizer import JsonDataGeneralizer
from schema_automator.importers.jsonschema_import_engine import JsonSchemaImportEngine
from schema_automator.importers.owl_import_engine import OwlImportEngine
from schema_automator.generalizers.rdf_data_generalizer import RdfDataGeneralizer
from schema_automator.importers.rdfs_import_engine import RdfsImportEngine
from schema_automator.importers.sql_import_engine import SqlImportEngine
from schema_automator.importers.tabular_import_engine import TableImportEngine
from schema_automator.utils.schemautils import write_schema

input_option = click.option(
    "-i",
    "--input",
    help="path to input file"
)
output_option = click.option(
    "-o",
    "--output",
    help="path to output file or directory."
)
schema_name_option = click.option(
    '--schema-name',
    '-n',
    default=DEFAULT_SCHEMA_NAME,
    show_default=True,
    help='Schema name')
schema_id_option = click.option(
    '--schema-id',
    help='Schema id')
annotator_option = click.option(
    '--annotator',
    '-A',
    help='name of annotator to use for auto-annotating results. Must be an OAK selector')
use_attributes_option = click.option(
    "--use-attributes/--no-use-attributes",
    help="If true, use attributes over slots/slot_usage"
)
column_separator_option = click.option('--column-separator', '-s', default='\t', help='separator')

# generalizer options

downcase_header_option = click.option('--downcase-header/--no-downcase-header', default=False, help='if true make headers lowercase')
snakecase_header_option = click.option('--snakecase-header/--no-snakecase-header', default=False, help='if true make headers snakecase')
infer_foreign_keys_option = click.option('--infer-foreign-keys/--no-infer-foreign-keys', default=False, help='infer ranges/foreign keys')
enum_columns_option = click.option('--enum-columns', '-E', multiple=True, help='column(s) that is forced to be an enum')
enum_mask_columns_option = click.option('--enum-mask-columns', multiple=True, help='column(s) that are excluded from being enums')
max_enum_size_option = click.option('--max-enum-size', default=50, help='do not create an enum if more than max distinct members')
enum_threshold_option = click.option('--enum-threshold', default=0.1, help='if the number of distinct values / rows is less than this, do not make an enum')


@click.group()
@click.option("-v", "--verbose",
              count=True,
              help="Set the level of verbosity")
@click.option("-q", "--quiet",
              help="Silence all diagnostics")
def main(verbose: int, quiet: bool):
    """Run the LinkML Schema Automator Command Line.

    A subcommand must be passed, for example:

        schemauto SUBCOMMAND [OPTIONS] ARGUMENTS

    To see logging or debugging info, the verbosity
    flag should be specified BEFORE the subcommand:

        schemauto -vv SUBCOMMAND [OPTIONS] ARGUMENTS
    """
    logger = logging.getLogger()
    if verbose >= 2:
        logger.setLevel(logging.DEBUG)
    elif verbose == 1:
        logger.setLevel(logging.INFO)
    else:
        logger.setLevel(logging.WARNING)
    if quiet:
        logger.setLevel(logging.ERROR)
    logging.info(f"Log level={verbose}")


@main.command()
@click.argument('tsvfile')  # input TSV (must have column headers
@output_option
@schema_name_option
@annotator_option
@click.option('--class-name', '-c', default=DEFAULT_CLASS_NAME, help='Core class name in schema')
@column_separator_option
@downcase_header_option
@snakecase_header_option
@enum_columns_option
@enum_threshold_option
@max_enum_size_option
@click.option('--data-dictionary-row-count',
              type=click.INT,
              help='rows that provide metadata about columns')
@click.option('--robot/--no-robot', default=False, help='set if the TSV is a ROBOT template')
@click.option('--pandera/--no-pandera', default=False, help='set to use panderas as inference engine')
def generalize_tsv(tsvfile, output, class_name, schema_name, pandera: bool, annotator, **kwargs):
    """
    Generalizes from a single TSV file to a single-class schema

    See :ref:`generalizers` for more on the generalization framework

    Example:

        schemauto generalize-tsv --class-name Person --schema-name PersonInfo my/data/persons.tsv
    """
    kwargs = {k:v for k, v in kwargs.items() if v is not None}
    if pandera:
        ie = PandasDataGeneralizer(**kwargs)
    else:
        ie = CsvDataGeneralizer(**kwargs)
    schema = ie.convert(tsvfile, class_name=class_name, schema_name=schema_name)
    if annotator:
        impl = get_implementation_from_shorthand(annotator)
        sa = SchemaAnnotator(impl)
        schema = sa.annotate_schema(schema)
    write_schema(schema, output)


@main.command()
@click.argument('tsvfiles', nargs=-1)  # input TSV (must have column headers
@output_option
@schema_name_option
@column_separator_option
@downcase_header_option
@snakecase_header_option
@enum_columns_option
@enum_threshold_option
@max_enum_size_option
@click.option('--robot/--no-robot', default=False, help='set if the TSV is a ROBOT template')
def generalize_tsvs(tsvfiles, output, schema_name, **kwargs):
    """
    Generalizes from a multiple TSV files to a multi-class schema

    See :ref:`generalizers` for more on the generalization framework

    This uses :ref:`CsvDataGeneralizer.convert_multiple`

    Example:

        schemauto generalize-tsvs --class-name Person --schema-name PersonInfo my/data/*.tsv
    """
    ie = CsvDataGeneralizer(**kwargs)
    schema = ie.convert_multiple(tsvfiles, schema_name=schema_name)
    write_schema(schema, output)


@main.command()
@click.argument('url')  # input TSV (must have column headers
@output_option
@schema_name_option
@column_separator_option
@downcase_header_option
@snakecase_header_option
@enum_columns_option
@enum_threshold_option
@max_enum_size_option
@click.option('--class-name', '-c', default=DEFAULT_CLASS_NAME, help='Core class name in schema')
@click.option('--pandera/--no-pandera', default=False, help='set to use panderas as inference engine')
@click.option('--data-output', help='Path to file of downloaded data')
@click.option('--table-number',
              type=int,
              default=0,
              show_default=True,
              help='If URL has multiple tables, use this one (zero-based)')
def generalize_htmltable(url, output, class_name, schema_name, pandera: bool,
                         table_number: int, data_output,
                         **kwargs):
    """
    Generalizes from a table parsed from a URL

    Uses pandas/beautiful soup.

    Note: if the website cannot be accessed directly, you can download the HTML
    and pass in an argument of the form file:///absolute/path/to/file.html
    """
    dfs = pd.read_html(url)
    logging.info(f"{url} has {len(dfs)} tables")
    df = dfs[table_number]
    if data_output:
        df.to_csv(data_output, index=False, sep="\t")
    if pandera:
        ge = PandasDataGeneralizer(**kwargs)
    else:
        ge = CsvDataGeneralizer(**kwargs)
    schema = ge.convert_from_dataframe(df, class_name=class_name, schema_name=schema_name)
    write_schema(schema, output)


@main.command()
@click.argument('dpfiles', nargs=-1) ## input DOSDPs
@output_option
@schema_name_option
@click.option('--range-as-enums/--no-range-as-enums',
              default=True,
              help="Model range ontology classes as enums")
def import_dosdps(dpfiles, output, **args):
    """
    Imports DOSDP pattern YAML to a LinkML schema

    See :ref:`importers` for more on the importers framework

    Example:

        schemauto import-dosdps --range-as-enums patterns/*yaml -o my-schema.yaml
    """
    ie = DOSDPImportEngine()
    schema = ie.convert(dpfiles, **args)
    write_schema(schema, output)


@main.command()
@click.argument('db')
@output_option
@schema_name_option
def import_sql(db, output, **args):
    """
    Imports a schema by introspecting a relational database

    See :ref:`importers` for more on the importers framework
    """
    ie = SqlImportEngine()
    schema = ie.convert(db, **args)
    write_schema(schema, output)


@main.command()
@output_option
@schema_name_option
@click.option('--class-name', '-c', default=DEFAULT_CLASS_NAME, help='Core class name in schema')
@click.option('--data-output', help='Path to file of downloaded data')
@click.option('--element-type', help='E.g. class, enum')
@click.option('--parent', help='parent ID')
@click.option('--columns',
              required=True,
              help='comma-separated schemasheets descriptors of each column. Must be in same order')
@click.option('--table-number',
              type=int,
              default=0,
              show_default=True,
              help='If URL has multiple tables, use this one (zero-based)')
@click.argument('url')  # input TSV (must have column headers
def import_htmltable(url, output, class_name, schema_name, columns,
                         table_number: int, data_output,
                         **kwargs):
    """
    Imports from a table parsed from a URL using SchemaSheets

    Uses pandas/beautiful soup
    """
    dfs = pd.read_html(url)
    logging.info(f"{url} has {len(dfs)} tables")
    df = dfs[table_number]
    if data_output:
        df.to_csv(data_output, index=False, sep="\t")
    ie = TableImportEngine(columns=columns.split(","), **kwargs)
    schema = ie.import_from_dataframe(df)
    write_schema(schema, output)


@main.command()
@click.argument('input')
@output_option
@schema_name_option
@click.option('--container-class-name', default='Container', help="name of root class")
@click.option('--format', '-f', default='json', help="json or yaml (or json.gz or yaml.gz) or frontmatter")
@click.option('--enum-columns', '-E', multiple=True, help='column(s) that is forced to be an enum')
@click.option('--enum-mask-columns', multiple=True, help='column(s) that are excluded from being enums')
@click.option('--max-enum-size', default=50, help='do not create an enum if more than max distinct members')
@click.option('--enum-threshold', default=0.1, help='if the number of distinct values / rows is less than this, do not make an enum')
@click.option('--omit-null/--no-omit-null', default=False, help="if true, ignore null values")
@click.option('--inlined-map', multiple=True, help="SLOT_NAME.KEY pairs indicating which slots are inlined as dict")
@click.option('--depluralize/--no-depluralized',
              default=True,
              show_default=True,
              help="Auto-depluralize class names to singular form")
def generalize_json(input, output, schema_name, depluralize: bool, format, omit_null, inlined_map, **kwargs):
    """
    Generalizes from a JSON file to a schema

    See :ref:`generalizers` for more on the generalization framework

    Example:

        schemauto generalize-json my/data/persons.json -o my.yaml
    """
    ie = JsonDataGeneralizer(omit_null=omit_null, depluralize_class_names=depluralize)
    if inlined_map:
        ie.inline_as_dict_slot_keys = dict([tuple(x.split(".")) for x in inlined_map])
    schema = ie.convert(input, format=format, **kwargs)
    write_schema(schema, output)


@main.command()
@click.argument('input')
@output_option
@schema_name_option
@click.option('--container-class-name', default='Container', help="name of root class")
@click.option('--enum-columns', '-E', multiple=True, help='column(s) that is forced to be an enum')
@click.option('--enum-mask-columns', multiple=True, help='column(s) that are excluded from being enums')
@click.option('--max-enum-size', default=50, help='do not create an enum if more than max distinct members')
@click.option('--enum-threshold', default=0.1, help='if the number of distinct values / rows is less than this, do not make an enum')
@click.option('--omit-null/--no-omit-null', default=False, help="if true, ignore null values")
def generalize_toml(input, output, schema_name, omit_null, **kwargs):
    """
    Generalizes from a TOML file to a schema

    See :ref:`generalizers` for more on the generalization framework

    Example:

        schemauto generalize-toml my/data/conf.toml -o my.yaml
    """
    ie = JsonDataGeneralizer(omit_null=omit_null)
    schema = ie.convert(input, format='toml', **kwargs)
    write_schema(schema, output)


@main.command()
@click.argument('input')
@output_option
@schema_name_option
@use_attributes_option
@click.option("--import-project/--no-import-project",
              help="If true, then the input path should be a directory with multiple schema files")
@click.option('--format', '-f', default='json', help='JSON Schema format - yaml or json')
def import_json_schema(input, output, import_project: bool, schema_name, format, **kwargs):
    """
    Imports from JSON Schema to LinkML

    See :ref:`importers` for more on the importer framework

    Example:

        schemauto import-json-schema my/schema/personinfo.schema.json
    """
    ie = JsonSchemaImportEngine(**kwargs)
    if not import_project:
        schema = ie.convert(input, name=schema_name, format=format)
        write_schema(schema, output)
    else:
        if output is None:
            raise ValueError(f"You must pass an export directory with --output")
        ie.import_project(input, output, name=schema_name, format=format)


@main.command()
@click.argument('input')
@output_option
@schema_name_option
@schema_id_option
def import_frictionless(input, output, schema_name, schema_id, **kwargs):
    """
    Imports from Frictionless data package to LinkML

    See :ref:`importers` for more on the importer framework

    Example:

        schemauto import-frictionless cfde.package.json
    """
    ie = FrictionlessImportEngine(**kwargs)
    schema = ie.convert(input, name=schema_name, id=schema_id)
    write_schema(schema, output)


@main.command()
@click.argument('owlfile')
@output_option
@schema_name_option
@click.option('--identifier', '-I', help="Slot to use as identifier")
@click.option('--model-uri', help="Model URI prefix")
@click.option('--output', '-o', help="Path to saved yaml schema")
def import_owl(owlfile, output, **args):
    """
    Import an OWL ontology to LinkML

    Note:
         - this works best for "schema-style" ontologies
         - input must be in functional syntax

    See :ref:`importers` for more on the importer framework

    Example:

        schemauto import-owl prov.ofn -o my.yaml
    """
    sie = OwlImportEngine()
    schema = sie.convert(owlfile, **args)
    write_schema(schema, output)


@main.command()
@click.argument('rdfsfile')
@output_option
@schema_name_option
@click.option('--input-type', '-I',
              default='turtle',
              help="Input format, eg. turtle")
@click.option('--identifier', '-I', help="Slot to use as identifier")
@click.option('--model-uri', help="Model URI prefix")
@click.option('--metamodel-mappings',
              help="Path to metamodel mappings YAML dicitonary")
@click.option('--output', '-o', help="Path to saved yaml schema")
def import_rdfs(rdfsfile, output, metamodel_mappings, **args):
    """
    Import an RDFS schema to LinkML

    Example:

        schemauto import-rdfs prov.rdfs.ttl -o prov.yaml
    """
    mappings_obj = None
    if metamodel_mappings:
        with open(metamodel_mappings) as f:
            mappings_obj = yaml.safe_load(f)
    sie = RdfsImportEngine(initial_metamodel_mappings=mappings_obj)
    schema = sie.convert(rdfsfile, **args)
    write_schema(schema, output)

@main.command()
@click.argument('rdffile')
@output_option
@click.option('--dir', '-d', required=True)
def generalize_rdf(rdffile, dir, output, **args):
    """
    Generalizes from an RDF file to a schema

    See :ref:`generalizers` for more on the generalization framework

    The input must be in turtle

    Example:

        schemauto generalize-rdf my/data/persons.ttl
    """
    sie = RdfDataGeneralizer()
    if not os.path.exists(dir):
        os.makedirs(dir)
    schema_dict = sie.convert(rdffile, dir=dir, format='ttl')
    write_schema(schema_dict, output)


@main.command()
@click.argument('schema')
@click.option('--curie-only/--no-curie-only',
              default=False,
              show_default=True,
              help="if set, only use results that are mapped to CURIEs")
@click.option('--input',
              '-i',
              help="OAK input ontology selector")
@output_option
def annotate_schema(schema: str, input: str, output: str, **kwargs):
    """
    Annotate all elements of a schema.

    This uses OAK (https://incatools.github.io/ontology-access-kit),
    and you can provide any OAK backend that supports text annotation.
    
    At this time, the best choice is likely the bioportal backend
    
    Example:
    
        schemauto annotate-schema -i bioportal: my-schema.yaml -o annotated.yaml
        
    This will require you setting the API key via OAK - see OAK docs.
        
    You can specify a specific ontology
    
        schemauto annotate-schema -i bioportal:ncbitaxon my-schema.yaml -o annotated.yaml

    In future OAK will support a much wider variety of annotators including:
    
       - OLS
       - SciSpacy
       - NLTK
       - OGER

    To see all possible selectors, see the OAK docs:
    
      - https://incatools.github.io/ontology-access-kit/selectors.html
    """
    impl = get_implementation_from_shorthand(input)
    annr = SchemaAnnotator(impl, **kwargs)
    schema = annr.annotate_schema(schema)
    write_schema(schema, output)

@main.command()
#@main.command()
@click.argument('schema')
@click.option('--input', '-i', help="OAK input ontology selector")
@click.option('--annotate/--no-annotate', default=True, help="If true, annotate the schema")
@output_option
def enrich_using_ontology(schema: str, input: str, output: str, annotate: bool, **args):
    """
    Enrich a schema using an ontology.

    Here, "enrich" means copying over metadata from the ontology to the schema.
    For example, if the schema has a class "Gene" that is mapped to a SO class for "gene",
    then calling this command will copy the SO class definition to the schema class.

    This will use OAK to add additional metadata using uris and mappings in the schema.

    See the OAK docs for options for which annotators to use; examples include:

     - bioportal:   # (include the colon) any ontology in bioportal

     - bioportal:umls   # a specific ontology in bioportal

     - my.obo       # any local OBO file

     - sqlite:obo:cl   # a specific OBO file or semsql registered ontology

    For example, if your schema has a class with a mapping to a SO class,
    then the definition of that will be copied to the class description.

    Example:

        schemauto enrich-using-ontology -i bioportal: my-schema.yaml -o my-enriched.yaml

    If your schema has no mappings you can use --annotate to add them

    Example:

        schemauto enrich-using-ontology -i so.obo --annotate my-schema.yaml -o my-enriched.yaml --annotate
    """
    impl = get_implementation_from_shorthand(input)
    annr = SchemaAnnotator(impl)
    logging.info(f"Enriching: {schema}")
    if annotate:
        schema = annr.annotate_schema(schema)
    schema = annr.enrich(schema)
    write_schema(schema, output)


@main.command()
@click.option('--model', '-m', help="Name of model")
@output_option
@click.argument('schema')
def enrich_using_llm(schema: str, model: str, output: str, **args):
    """
    Enrich a schema using an LLM.

    Example:

        schemauto enrich-using-llm -m gpt-4-turbo my-schema.yaml -o my-enriched.yaml

    This will enrich the schema by adding missing description fields. In future
    other enrichments may be possible.

    Note for this to work, you will need to have LLM installed as an extra.

    Example:

        pip install schema-automator[llm]

    """
    logging.info(f"Enriching: {schema}")
    schema = llm_annotator.enrich_using_llm(schema, model)
    write_schema(schema, output)


@main.command()
@click.argument('schema')
@output_option
def annotate_using_jsonld(schema: str, output: str, **args):
    """
    Annotates a schema using a Json-LD context file
    """
    annr = JsonLdAnnotator()
    schemadef = SchemaDefinition(schema)
    annr.annotate_schema(schemadef)
    write_schema(schemadef, output)


if __name__ == "__main__":
    main()



import logging

import click
from typing import Union, Dict, List, Any, Mapping, Collection
from collections import defaultdict
import json

import tomlkit
import yaml
import gzip

from dataclasses import dataclass, field

from linkml_runtime import SchemaView
from linkml_runtime.linkml_model import SchemaDefinition, SlotDefinitionName

from schema_automator.generalizers.generalizer import Generalizer
from schema_automator.generalizers.csv_data_generalizer import CsvDataGeneralizer
from linkml_runtime.utils.formatutils import camelcase

from schema_automator.utils.schemautils import write_schema


@dataclass
class JsonDataGeneralizer(Generalizer):
    """
    A generalizer that abstract from JSON instance data
    """
    mappings: dict = None
    omit_null: bool = None

    inline_as_dict_slot_keys: Mapping[str, str] = None
    """Mapping between the name of a dict-inlined slot and the unique key for that entity
    """



    def convert(self, input: Union[str, Dict], format: str = 'json',
                container_class_name='Container',
                **kwargs) -> SchemaDefinition:
        """
        Generalizes from a JSON file

        :param input:
        :param format:
        :param container_class_name:
        :param kwargs:
        :return:
        """
        csv_engine = CsvDataGeneralizer(**kwargs)

        if isinstance(input, dict):
            obj = input
        else:
            if format.endswith('.gz'):
                format = format.replace('.gz', '')
                stream = gzip.open(input)
            else:
                stream = open(input)
            with stream:
                if format == 'json':
                    obj = json.load(stream)
                elif format == 'yaml':
                    obj = yaml.safe_load(stream)
                elif format == 'toml':
                    obj_str = "".join(stream.readlines())
                    toml_obj = tomlkit.parse(obj_str)
                    json_str = json.dumps(toml_obj)
                    obj = json.loads(json_str)
                else:
                    raise Exception(f'bad format {format}')
        rows_by_table = defaultdict(list)
        self.rows_by_table = rows_by_table
        self._convert_obj(obj, table=container_class_name)
        schemas = []
        for cn, rows_dict in rows_by_table.items():
            schema = csv_engine.convert_dicts(rows_dict, cn, cn)
            schemas.append(schema)
        sv = SchemaView(schemas[0])
        for s in schemas[1:]:
            sv.merge_schema(s)
        schema = sv.schema
        schema.classes[container_class_name].tree_root = True
        self.add_additional_info(schema)
        return schema

    def _key_to_classname(self, k: str) -> str:
        return camelcase(k)

    def _convert_obj(self, obj, table='Container'):
        """
        Recursively transform an object into flattened key-value lists

        :param obj:
        :param table:
        :return:
        """
        if isinstance(obj, dict):
            row = defaultdict(set)
            for k, v in obj.items():
                if v is None and self.omit_null:
                    continue
                if self.inline_as_dict_slot_keys and k in self.inline_as_dict_slot_keys:
                    key_name = self.inline_as_dict_slot_keys[SlotDefinitionName(k)]
                    self.identifier_slots.append(key_name)
                    #print(f"INLINED: {key_name} = {v}")
                    v = self._inlined_dict_to_list(v, key_name)
                tbl_name = k
                if self.depluralize_class_names:
                    singular_noun = self.inflect_engine.singular_noun(tbl_name)
                    if singular_noun:
                        logging.info(f"Depluralized: {tbl_name} => {singular_noun}")
                        tbl_name = singular_noun
                tbl_name = camelcase(tbl_name)
                row[k] = self._convert_obj(v, table=tbl_name)
            self.rows_by_table[table].append(row)
            return f'$ref:{table}'
        elif isinstance(obj, list):
            new_list = [self._convert_obj(v, table=table) for v in obj]
            return new_list
        else:
            return obj

    def _inlined_dict_to_list(self, inlined_dict: Dict[str, dict], key_name: str) -> list:
        rows = []
        for k, v in inlined_dict.items():
            if isinstance(v, dict):
                rows.append({**v, key_name: k})
            elif isinstance(v, list):
                raise ValueError(f"Cannot handle an inlined dict of form {inlined_dict} for key={k}")
            else:
                rows.append({key_name: k, f"{key_name}_value": v})
        return rows



    def _as_name(self, v):
        v = str(v)
        for sep in ['#', '/']:
            if sep in v:
                return v.split(sep)[-1]
        return v

def parse_frontmatter_files(paths: List[str], text_slot='_text') -> Any:
    blocks = []
    for path in paths:
        with open(path) as stream:
            state = 0
            yamlstr = ""
            txt = ""
            for line in stream.readlines():
                if line.startswith('---'):
                    state += 1
                else:
                    if state == 1:
                        yamlstr += line
                    elif state == 2:
                        txt += line
                    elif state > 2:
                        raise Exception(f'Limited to one frontmatter block per file')
            obj = yaml.safe_load(yamlstr)
            obj[text_slot] = txt
            blocks.append(obj)
    return blocks



@click.command()
@click.argument('input')
@click.option('--container-class-name', help="name of root class")
@click.option('--format', '-f', default='json', help="json or yaml (or json.gz or yaml.gz) or frontmatter")
@click.option('--enum-columns', '-E', multiple=True, help='column(s) that is forced to be an enum')
@click.option('--enum-mask-columns', multiple=True, help='column(s) that are excluded from being enums')
@click.option('--max-enum-size', default=50, help='do not create an enum if more than max distinct members')
@click.option('--enum-threshold', default=0.1, help='if the number of distinct values / rows is less than this, do not make an enum')
@click.option('--omit-null/--no-omit-null', default=False, help="if true, ignore null values")
def json2model(input, format, omit_null, **kwargs):
    """ Infer a model from JSON instance data


    """
    ie = JsonDataGeneralizer(omit_null=omit_null)
    schema = ie.convert(input, dir=dir, format=format, **kwargs)
    write_schema(schema)

@click.command()
@click.argument('inputs', nargs=-1)
@click.option('--container-class-name', help="name of root class")
@click.option('--enum-columns', '-E', multiple=True, help='column(s) that is forced to be an enum')
@click.option('--enum-mask-columns', multiple=True, help='column(s) that are excluded from being enums')
@click.option('--max-enum-size', default=50, help='do not create an enum if more than max distinct members')
@click.option('--enum-threshold', default=0.1, help='if the number of distinct values / rows is less than this, do not make an enum')
@click.option('--omit-null/--no-omit-null', default=False, help="if true, ignore null values")
def frontmatter2model(inputs, format, omit_null, **kwargs):
    """ Infer a model from frontmatter files


    """
    ie = JsonDataGeneralizer(omit_null=omit_null)
    objs = parse_frontmatter_files(list(inputs))
    schema = ie.convert({'objects': objs}, dir=dir, format=format, **kwargs)
    write_schema(schema)

if __name__ == '__main__':
    json2model()





import click
from typing import Union, Dict, List, Any
from collections import defaultdict
import json
import yaml
import gzip

from dataclasses import dataclass

from linkml_runtime import SchemaView
from linkml_runtime.linkml_model import SchemaDefinition

from schema_automator import JsonDataGeneralizer
from schema_automator.generalizers.generalizer import Generalizer
from schema_automator.generalizers.csv_data_generalizer import CsvDataGeneralizer
from linkml_runtime.utils.formatutils import camelcase

from schema_automator.utils.schemautils import write_schema


@dataclass
class TomlDataGeneralizer(Generalizer):
    """
    A generalizer that abstract from TOML instance data
    """
    mappings: dict = None
    omit_null: bool = None

    def convert(self, input: str, **kwargs) -> SchemaDefinition:
        """
        Generalizes from a JSON file

        :param input:
        :param kwargs:
        :return:
        """
        w
        json_engine = JsonDataGeneralizer(**kwargs)
        return json_engine.convert(input, format='toml', **kwargs)


from abc import ABC
from collections.abc import Collection, Set
from dataclasses import dataclass, field
from typing import List

import inflect
from linkml_runtime.linkml_model import SchemaDefinition, Prefix

DEFAULT_CLASS_NAME = 'Observation'
DEFAULT_SCHEMA_NAME = 'MySchema'


@dataclass
class Generalizer(ABC):
    """
    Abstract Base Class for all Generalization Engines.

    Generalization Engines take example data in some format and generalizes to a new SchemaDefinition
    """

    identifier_slots: List[str] = field(default_factory=lambda: [])
    """Slots asserted to be identifiers"""

    depluralize_class_names: bool = field(default_factory=lambda: False)

    inflect_engine: inflect.engine = field(default_factory=lambda: inflect.engine())


    def convert(self, file: str, **kwargs) -> SchemaDefinition:
        raise NotImplementedError

    def add_prefix(self, schema: SchemaDefinition, prefix: str, url: str):
        schema.prefixes[prefix] = Prefix(prefix, url)

    def add_default_prefixes(self, schema: SchemaDefinition):
        self.add_prefix(schema, 'linkml', 'https://w3id.org/linkml/')

    def add_additional_info(self, schema: SchemaDefinition) -> None:
        for s in self.identifier_slots:
            schema.slots[s].identifier = True


import click
from typing import Union, Dict, List, Any, Type

import pandas as pd
import pandera as pa
from dataclasses import dataclass

from linkml.utils.schema_builder import SchemaBuilder
from linkml.utils.schema_fixer import SchemaFixer
from linkml_runtime.linkml_model import SchemaDefinition, ClassDefinition, SlotDefinition
from pandera import Column

from schema_automator import CsvDataGeneralizer
from schema_automator.generalizers.generalizer import DEFAULT_CLASS_NAME

DATA_TYPE_MAP = {
    pa.Int: 'integer',
    pa.Float: 'float',
    pa.String: 'string',
    pa.DateTime: 'datetime',
}

def map_dtype(t: Type) -> str:
    for type_cls, rng in DATA_TYPE_MAP.items():
        if isinstance(t, type_cls):
            return rng
    return 'string'


@dataclass
class PandasDataGeneralizer(CsvDataGeneralizer):
    """
    A generalizer that abstracts from Pandas data frames
    """

    def convert(self, input: Union[pd.DataFrame, str], format: str = 'tsv',
                container_class_name='Container',
                class_name=DEFAULT_CLASS_NAME,
                **kwargs) -> SchemaDefinition:
        if isinstance(input, str):
            df = pd.read_csv(input, sep=self.column_separator)
        elif isinstance(input, pd.DataFrame):
            df = input
        else:
            raise ValueError(f"{input} must be DataFrame or path to TSV")
        c = self.from_dataframe(df, class_name=class_name)
        sb = SchemaBuilder(self.schema_name)
        sb.add_defaults()
        schema = sb.schema
        schema.classes[c.name] = c
        sf = SchemaFixer()
        sf.add_container(schema, class_name=container_class_name)
        return schema

    def from_dataframe(self, df: pd.DataFrame, class_name=DEFAULT_CLASS_NAME) -> ClassDefinition:
        ps = pa.infer_schema(df)
        return self.from_dataframe_schema(ps, class_name=class_name)

    def from_dataframe_schema(self, ps: pa.DataFrameSchema, class_name=DEFAULT_CLASS_NAME) -> ClassDefinition:
        if ps.name:
            name = ps.name
        else:
            name = class_name
        c = ClassDefinition(name)
        for col_name, col in ps.columns.items():
            slot = self.from_column(col_name, col)
            c.attributes[slot.name] = slot
        return c

    def from_column(self, col_name: str, col: Column) -> SlotDefinition:
        return SlotDefinition(col_name,
                              range=map_dtype(col.dtype))




import click
import logging
import yaml
from typing import Union, Dict, Tuple, List, Any
from collections import defaultdict

from linkml.utils.schema_builder import SchemaBuilder
from linkml_runtime.linkml_model import (
    SchemaDefinition,
    SlotDefinition,
    ClassDefinition,
)
from funowl.converters.functional_converter import to_python
from funowl import *

from dataclasses import dataclass, field

from linkml_runtime.utils.introspection import package_schemaview
from rdflib import Graph, RDF, OWL, URIRef, RDFS, SKOS, SDO, Namespace
from schema_automator.importers.import_engine import ImportEngine
from schema_automator.utils.schemautils import write_schema


HTTP_SDO = Namespace("http://schema.org/")

DEFAULT_METAMODEL_MAPPINGS = {
    "is_a": [RDFS.subClassOf, SKOS.broader],
    "domain_of": [HTTP_SDO.domainIncludes, SDO.domainIncludes],
    "rangeIncludes": [HTTP_SDO.rangeIncludes, SDO.rangeIncludes],
    "exact_mappings": [OWL.sameAs, HTTP_SDO.sameAs],
    ClassDefinition.__name__: [RDFS.Class, OWL.Class, SKOS.Concept],
    SlotDefinition.__name__: [
        RDF.Property,
        OWL.ObjectProperty,
        OWL.DatatypeProperty,
        OWL.AnnotationProperty,
    ],
}


@dataclass
class RdfsImportEngine(ImportEngine):
    """
    An ImportEngine that takes RDFS and converts it to a LinkML schema
    """

    mappings: dict = None
    initial_metamodel_mappings: Dict[str, List[URIRef]] = None
    metamodel_mappings: Dict[str, List[URIRef]] = None
    reverse_metamodel_mappings: Dict[URIRef, List[str]] = None
    include_unmapped_annotations = False
    metamodel = None

    def __post_init__(self):
        sv = package_schemaview("linkml_runtime.linkml_model.meta")
        self.metamodel = sv
        self.metamodel_mappings = defaultdict(list)
        self.reverse_metamodel_mappings = defaultdict(list)
        for k, vs in DEFAULT_METAMODEL_MAPPINGS.items():
            self.metamodel_mappings[k].extend(vs)
            for v in vs:
                self.reverse_metamodel_mappings[v].append(k)
        if self.initial_metamodel_mappings:
            for k, vs in self.initial_metamodel_mappings.items():
                self.metamodel_mappings[k].extend(vs)
                for v in vs:
                    self.reverse_metamodel_mappings[v].append(k)
        for e in sv.all_elements().values():
            mappings = []
            for ms in sv.get_mappings(e.name, expand=True).values():
                for m in ms:
                    uri = URIRef(m)
                    mappings.append(uri)
                    self.reverse_metamodel_mappings[uri].append(e.name)
            self.metamodel_mappings[e.name] = mappings

    def convert(
        self,
        file: str,
        name: str = None,
        format="turtle",
        default_prefix: str = None,
        model_uri: str = None,
        identifier: str = None,
        **kwargs,
    ) -> SchemaDefinition:
        """
        Converts an OWL schema-style ontology

        :param file:
        :param name:
        :param model_uri:
        :param identifier:
        :param kwargs:
        :return:
        """
        self.mappings = {}
        g = Graph()
        g.parse(file, format=format)
        if name is not None and default_prefix is None:
            default_prefix = name
        if name is None:
            name = default_prefix
        if name is None:
            name = "example"
        sb = SchemaBuilder(name=name)
        sb.add_defaults()
        schema = sb.schema
        for k, v in g.namespaces():
            if k == "schema" and v != "http://schema.org/":
                continue
            sb.add_prefix(k, v, replace_if_present=True)
        if default_prefix is not None:
            schema.default_prefix = default_prefix
            if default_prefix not in schema.prefixes:
                sb.add_prefix(default_prefix, model_uri, replace_if_present=True)
            schema.id = schema.prefixes[default_prefix].prefix_reference
        cls_slots = defaultdict(list)
        props = []
        for rdfs_property_metaclass in self._rdfs_metamodel_iri(
            SlotDefinition.__name__
        ):
            for p in g.subjects(RDF.type, rdfs_property_metaclass):
                props.append(p)
        # implicit properties
        for metap in (
            self.reverse_metamodel_mappings["domain_of"]
            + self.reverse_metamodel_mappings["rangeIncludes"]
        ):
            for p, _, _o in g.triples((None, metap, None)):
                props.append(p)
        for p in set(props):
            sn = self.iri_to_name(p)
            init_dict = self._dict_for_subject(g, p)
            if "domain_of" in init_dict:
                for x in init_dict["domain_of"]:
                    cls_slots[x].append(sn)
                del init_dict["domain_of"]
            if "rangeIncludes" in init_dict:
                init_dict["any_of"] = [{"range": x} for x in init_dict["rangeIncludes"]]
                del init_dict["rangeIncludes"]
            slot = SlotDefinition(sn, **init_dict)
            slot.slot_uri = str(p.n3(g.namespace_manager))
            sb.add_slot(slot)
        rdfs_classes = []
        for rdfs_class_metaclass in self._rdfs_metamodel_iri(ClassDefinition.__name__):
            for s in g.subjects(RDF.type, rdfs_class_metaclass):
                rdfs_classes.append(s)
        # implicit classes
        for metap in [RDFS.subClassOf]:
            for s, _, o in g.triples((None, metap, None)):
                rdfs_classes.append(s)
                rdfs_classes.append(o)
        for s in set(rdfs_classes):
            cn = self.iri_to_name(s)
            init_dict = self._dict_for_subject(g, s)
            c = ClassDefinition(cn, **init_dict)
            c.slots = cls_slots.get(cn, [])
            c.class_uri = str(s.n3(g.namespace_manager))
            sb.add_class(c)
        if identifier is not None:
            id_slot = SlotDefinition(identifier, identifier=True, range="uriorcurie")
            schema.slots[identifier] = id_slot
            for c in schema.classes.values():
                if not c.is_a and not c.mixins:
                    if identifier not in c.slots:
                        c.slots.append(identifier)
        return schema

    def _dict_for_subject(self, g: Graph, s: URIRef) -> Dict[str, Any]:
        """
        Looks up triples for a subject and converts to dict using linkml keys.

        :param g:
        :param p:
        :return:
        """
        init_dict = {}
        for pp, obj in g.predicate_objects(s):
            if pp == RDF.type:
                continue
            metaslot_name = self._element_from_iri(pp)
            if metaslot_name is None:
                logging.warning(f"Not mapping {pp}")
                continue
            if metaslot_name == "name":
                metaslot_name = "title"
            v = self._object_to_value(obj)
            metaslot = self.metamodel.get_slot(metaslot_name)
            if not metaslot or metaslot.multivalued:
                if metaslot_name not in init_dict:
                    init_dict[metaslot_name] = []
                init_dict[metaslot_name].append(v)
            else:
                init_dict[metaslot_name] = v
        return init_dict

    def _rdfs_metamodel_iri(self, name: str) -> List[URIRef]:
        return self.metamodel_mappings.get(name, [])

    def _element_from_iri(self, iri: URIRef) -> str:
        r = self.reverse_metamodel_mappings.get(iri, [])
        if len(r) > 0:
            if len(r) > 1:
                logging.info(f"Multiple mappings for {iri}: {r}")
            return r[0]

    def _object_to_value(self, obj: Any) -> Any:
        if isinstance(obj, URIRef):
            return self.iri_to_name(obj)
        if isinstance(obj, Literal):
            return obj.value
        return obj

    def iri_to_name(self, v: URIRef) -> str:
        n = self._as_name(v)
        if n != v:
            self.mappings[n] = v
        return n

    def _as_name(self, v: URIRef):
        v = str(v)
        for sep in ["#", "/", ":"]:
            if sep in v:
                return v.split(sep)[-1]
        return v



# Auto generated from frictionless.yaml by pythongen.py version: 0.9.0
# Generation date: 2022-07-08T14:42:16
# Schema: frictionless
#
# id: https://w3id.org/linkml/frictionless
# description: frictionless
# license: https://creativecommons.org/publicdomain/zero/1.0/

import dataclasses
import sys
import re
from jsonasobj2 import JsonObj, as_dict
from typing import Optional, List, Union, Dict, ClassVar, Any
from dataclasses import dataclass
from linkml_runtime.linkml_model.meta import EnumDefinition, PermissibleValue, PvFormulaOptions

from linkml_runtime.utils.slot import Slot
from linkml_runtime.utils.metamodelcore import empty_list, empty_dict, bnode
from linkml_runtime.utils.yamlutils import YAMLRoot, extended_str, extended_float, extended_int
from linkml_runtime.utils.dataclass_extensions_376 import dataclasses_init_fn_with_kwargs
from linkml_runtime.utils.formatutils import camelcase, underscore, sfx
from linkml_runtime.utils.enumerations import EnumDefinitionImpl
from rdflib import Namespace, URIRef
from linkml_runtime.utils.curienamespace import CurieNamespace
from linkml_runtime.linkml_model.types import Boolean, String
from linkml_runtime.utils.metamodelcore import Bool

metamodel_version = "1.7.0"
version = None

# Overwrite dataclasses _init_fn to add **kwargs in __init__
dataclasses._init_fn = dataclasses_init_fn_with_kwargs

# Namespaces
ANY = CurieNamespace('Any', 'http://example.org/UNKNOWN/Any/')
FRICTIONLESS = CurieNamespace('frictionless', 'https://w3id.org/linkml/frictionless/')
LINKML = CurieNamespace('linkml', 'https://w3id.org/linkml/')
DEFAULT_ = FRICTIONLESS


# Types

# Class references
class FieldName(extended_str):
    pass


class ResourceName(extended_str):
    pass


class PackageName(extended_str):
    pass


Any = Any

@dataclass
class Dialect(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = FRICTIONLESS.Dialect
    class_class_curie: ClassVar[str] = "frictionless:Dialect"
    class_name: ClassVar[str] = "Dialect"
    class_model_uri: ClassVar[URIRef] = FRICTIONLESS.Dialect

    delimiter: Optional[str] = None
    doubleQuote: Optional[Union[bool, Bool]] = None
    lineTerminator: Optional[str] = None
    skipInitialSpace: Optional[Union[bool, Bool]] = None
    header: Optional[Union[bool, Bool]] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self.delimiter is not None and not isinstance(self.delimiter, str):
            self.delimiter = str(self.delimiter)

        if self.doubleQuote is not None and not isinstance(self.doubleQuote, Bool):
            self.doubleQuote = Bool(self.doubleQuote)

        if self.lineTerminator is not None and not isinstance(self.lineTerminator, str):
            self.lineTerminator = str(self.lineTerminator)

        if self.skipInitialSpace is not None and not isinstance(self.skipInitialSpace, Bool):
            self.skipInitialSpace = Bool(self.skipInitialSpace)

        if self.header is not None and not isinstance(self.header, Bool):
            self.header = Bool(self.header)

        super().__post_init__(**kwargs)


@dataclass
class Constraints(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = FRICTIONLESS.Constraints
    class_class_curie: ClassVar[str] = "frictionless:Constraints"
    class_name: ClassVar[str] = "Constraints"
    class_model_uri: ClassVar[URIRef] = FRICTIONLESS.Constraints

    required: Optional[Union[bool, Bool]] = None
    pattern: Optional[str] = None
    unique: Optional[Union[bool, Bool]] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self.required is not None and not isinstance(self.required, Bool):
            self.required = Bool(self.required)

        if self.pattern is not None and not isinstance(self.pattern, str):
            self.pattern = str(self.pattern)

        if self.unique is not None and not isinstance(self.unique, Bool):
            self.unique = Bool(self.unique)

        super().__post_init__(**kwargs)


@dataclass
class Field(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = FRICTIONLESS.Field
    class_class_curie: ClassVar[str] = "frictionless:Field"
    class_name: ClassVar[str] = "Field"
    class_model_uri: ClassVar[URIRef] = FRICTIONLESS.Field

    name: Union[str, FieldName] = None
    description: Optional[Union[str, List[str]]] = empty_list()
    type: Optional[Union[str, "TypeEnum"]] = None
    constraints: Optional[Union[dict, Constraints]] = None
    format: Optional[Union[str, "FormatEnum"]] = None
    enum: Optional[Union[str, List[str]]] = empty_list()

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._is_empty(self.name):
            self.MissingRequiredField("name")
        if not isinstance(self.name, FieldName):
            self.name = FieldName(self.name)

        if not isinstance(self.description, list):
            self.description = [self.description] if self.description is not None else []
        self.description = [v if isinstance(v, str) else str(v) for v in self.description]

        if self.type is not None and not isinstance(self.type, TypeEnum):
            self.type = TypeEnum(self.type)

        if self.constraints is not None and not isinstance(self.constraints, Constraints):
            self.constraints = Constraints(**as_dict(self.constraints))

        if self.format is not None and not isinstance(self.format, FormatEnum):
            self.format = FormatEnum(self.format)

        if not isinstance(self.enum, list):
            self.enum = [self.enum] if self.enum is not None else []
        self.enum = [v if isinstance(v, str) else str(v) for v in self.enum]

        super().__post_init__(**kwargs)


@dataclass
class Reference(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = FRICTIONLESS.Reference
    class_class_curie: ClassVar[str] = "frictionless:Reference"
    class_name: ClassVar[str] = "Reference"
    class_model_uri: ClassVar[URIRef] = FRICTIONLESS.Reference

    resource: Optional[str] = None
    fields: Optional[Union[Union[str, FieldName], List[Union[str, FieldName]]]] = empty_list()
    range: Optional[str] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self.resource is not None and not isinstance(self.resource, str):
            self.resource = str(self.resource)

        if not isinstance(self.fields, list):
            self.fields = [self.fields] if self.fields is not None else []
        self.fields = [v if isinstance(v, FieldName) else FieldName(v) for v in self.fields]

        if self.range is not None and not isinstance(self.range, str):
            self.range = str(self.range)

        super().__post_init__(**kwargs)


@dataclass
class ForeignKeys(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = FRICTIONLESS.ForeignKeys
    class_class_curie: ClassVar[str] = "frictionless:ForeignKeys"
    class_name: ClassVar[str] = "ForeignKeys"
    class_model_uri: ClassVar[URIRef] = FRICTIONLESS.ForeignKeys

    fields: Optional[Union[Union[str, FieldName], List[Union[str, FieldName]]]] = empty_list()
    reference: Optional[Union[dict, Reference]] = None
    constraint_name: Optional[str] = None
    range: Optional[str] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if not isinstance(self.fields, list):
            self.fields = [self.fields] if self.fields is not None else []
        self.fields = [v if isinstance(v, FieldName) else FieldName(v) for v in self.fields]

        if self.reference is not None and not isinstance(self.reference, Reference):
            self.reference = Reference(**as_dict(self.reference))

        if self.constraint_name is not None and not isinstance(self.constraint_name, str):
            self.constraint_name = str(self.constraint_name)

        if self.range is not None and not isinstance(self.range, str):
            self.range = str(self.range)

        super().__post_init__(**kwargs)


@dataclass
class Schema(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = FRICTIONLESS.Schema
    class_class_curie: ClassVar[str] = "frictionless:Schema"
    class_name: ClassVar[str] = "Schema"
    class_model_uri: ClassVar[URIRef] = FRICTIONLESS.Schema

    fields: Optional[Union[Dict[Union[str, FieldName], Union[dict, Field]], List[Union[dict, Field]]]] = empty_dict()
    missingValues: Optional[Union[str, List[str]]] = empty_list()
    primaryKey: Optional[Union[Union[str, FieldName], List[Union[str, FieldName]]]] = empty_list()
    foreignKeys: Optional[Union[Union[dict, ForeignKeys], List[Union[dict, ForeignKeys]]]] = empty_list()

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        self._normalize_inlined_as_list(slot_name="fields", slot_type=Field, key_name="name", keyed=True)

        if not isinstance(self.missingValues, list):
            self.missingValues = [self.missingValues] if self.missingValues is not None else []
        self.missingValues = [v if isinstance(v, str) else str(v) for v in self.missingValues]

        if not isinstance(self.primaryKey, list):
            self.primaryKey = [self.primaryKey] if self.primaryKey is not None else []
        self.primaryKey = [v if isinstance(v, FieldName) else FieldName(v) for v in self.primaryKey]

        if not isinstance(self.foreignKeys, list):
            self.foreignKeys = [self.foreignKeys] if self.foreignKeys is not None else []
        self.foreignKeys = [v if isinstance(v, ForeignKeys) else ForeignKeys(**as_dict(v)) for v in self.foreignKeys]

        super().__post_init__(**kwargs)


@dataclass
class Resource(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = FRICTIONLESS.Resource
    class_class_curie: ClassVar[str] = "frictionless:Resource"
    class_name: ClassVar[str] = "Resource"
    class_model_uri: ClassVar[URIRef] = FRICTIONLESS.Resource

    name: Union[str, ResourceName] = None
    profile: Optional[Union[str, "ProfileEnum"]] = None
    title: Optional[str] = None
    path: Optional[str] = None
    dialect: Optional[str] = None
    description: Optional[Union[str, List[str]]] = empty_list()
    schema: Optional[Union[dict, Schema]] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._is_empty(self.name):
            self.MissingRequiredField("name")
        if not isinstance(self.name, ResourceName):
            self.name = ResourceName(self.name)

        if self.profile is not None and not isinstance(self.profile, ProfileEnum):
            self.profile = ProfileEnum(self.profile)

        if self.title is not None and not isinstance(self.title, str):
            self.title = str(self.title)

        if self.path is not None and not isinstance(self.path, str):
            self.path = str(self.path)

        if self.dialect is not None and not isinstance(self.dialect, str):
            self.dialect = str(self.dialect)

        if not isinstance(self.description, list):
            self.description = [self.description] if self.description is not None else []
        self.description = [v if isinstance(v, str) else str(v) for v in self.description]

        if self.schema is not None and not isinstance(self.schema, Schema):
            self.schema = Schema(**as_dict(self.schema))

        super().__post_init__(**kwargs)


@dataclass
class Package(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = FRICTIONLESS.Package
    class_class_curie: ClassVar[str] = "frictionless:Package"
    class_name: ClassVar[str] = "Package"
    class_model_uri: ClassVar[URIRef] = FRICTIONLESS.Package

    name: Union[str, PackageName] = None
    profile: Optional[Union[str, "ProfileEnum"]] = None
    title: Optional[str] = None
    resources: Optional[Union[Dict[Union[str, ResourceName], Union[dict, Resource]], List[Union[dict, Resource]]]] = empty_dict()

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._is_empty(self.name):
            self.MissingRequiredField("name")
        if not isinstance(self.name, PackageName):
            self.name = PackageName(self.name)

        if self.profile is not None and not isinstance(self.profile, ProfileEnum):
            self.profile = ProfileEnum(self.profile)

        if self.title is not None and not isinstance(self.title, str):
            self.title = str(self.title)

        self._normalize_inlined_as_list(slot_name="resources", slot_type=Resource, key_name="name", keyed=True)

        super().__post_init__(**kwargs)


# Enumerations
class TypeEnum(EnumDefinitionImpl):

    string = PermissibleValue(text="string",
                                   description="string")
    datetime = PermissibleValue(text="datetime",
                                       description="datetime")
    boolean = PermissibleValue(text="boolean",
                                     description="boolean")
    integer = PermissibleValue(text="integer",
                                     description="integer")
    array = PermissibleValue(text="array",
                                 description="array")
    number = PermissibleValue(text="number",
                                   description="number")

    _defn = EnumDefinition(
        name="TypeEnum",
    )

class FormatEnum(EnumDefinitionImpl):

    any = PermissibleValue(text="any",
                             description="any")
    email = PermissibleValue(text="email",
                                 description="email")
    binary = PermissibleValue(text="binary",
                                   description="binary")

    _defn = EnumDefinition(
        name="FormatEnum",
    )

class ProfileEnum(EnumDefinitionImpl):

    _defn = EnumDefinition(
        name="ProfileEnum",
    )

    @classmethod
    def _addvals(cls):
        setattr(cls, "tabular-data-resource",
                PermissibleValue(text="tabular-data-resource") )
        setattr(cls, "tabular-data-package",
                PermissibleValue(text="tabular-data-package") )

# Slots
class slots:
    pass

slots.delimiter = Slot(uri=FRICTIONLESS.delimiter, name="delimiter", curie=FRICTIONLESS.curie('delimiter'),
                   model_uri=FRICTIONLESS.delimiter, domain=None, range=Optional[str])

slots.doubleQuote = Slot(uri=FRICTIONLESS.doubleQuote, name="doubleQuote", curie=FRICTIONLESS.curie('doubleQuote'),
                   model_uri=FRICTIONLESS.doubleQuote, domain=None, range=Optional[Union[bool, Bool]])

slots.lineTerminator = Slot(uri=FRICTIONLESS.lineTerminator, name="lineTerminator", curie=FRICTIONLESS.curie('lineTerminator'),
                   model_uri=FRICTIONLESS.lineTerminator, domain=None, range=Optional[str])

slots.skipInitialSpace = Slot(uri=FRICTIONLESS.skipInitialSpace, name="skipInitialSpace", curie=FRICTIONLESS.curie('skipInitialSpace'),
                   model_uri=FRICTIONLESS.skipInitialSpace, domain=None, range=Optional[Union[bool, Bool]])

slots.header = Slot(uri=FRICTIONLESS.header, name="header", curie=FRICTIONLESS.curie('header'),
                   model_uri=FRICTIONLESS.header, domain=None, range=Optional[Union[bool, Bool]])

slots.required = Slot(uri=LINKML.required, name="required", curie=LINKML.curie('required'),
                   model_uri=FRICTIONLESS.required, domain=None, range=Optional[Union[bool, Bool]])

slots.pattern = Slot(uri=LINKML.pattern, name="pattern", curie=LINKML.curie('pattern'),
                   model_uri=FRICTIONLESS.pattern, domain=None, range=Optional[str])

slots.unique = Slot(uri=LINKML.identifier, name="unique", curie=LINKML.curie('identifier'),
                   model_uri=FRICTIONLESS.unique, domain=None, range=Optional[Union[bool, Bool]])

slots.name = Slot(uri=LINKML.name, name="name", curie=LINKML.curie('name'),
                   model_uri=FRICTIONLESS.name, domain=None, range=URIRef)

slots.description = Slot(uri=LINKML.description, name="description", curie=LINKML.curie('description'),
                   model_uri=FRICTIONLESS.description, domain=None, range=Optional[Union[str, List[str]]])

slots.type = Slot(uri=FRICTIONLESS.type, name="type", curie=FRICTIONLESS.curie('type'),
                   model_uri=FRICTIONLESS.type, domain=None, range=Optional[Union[str, "TypeEnum"]])

slots.constraints = Slot(uri=FRICTIONLESS.constraints, name="constraints", curie=FRICTIONLESS.curie('constraints'),
                   model_uri=FRICTIONLESS.constraints, domain=None, range=Optional[Union[dict, Constraints]])

slots.format = Slot(uri=FRICTIONLESS.format, name="format", curie=FRICTIONLESS.curie('format'),
                   model_uri=FRICTIONLESS.format, domain=None, range=Optional[Union[str, "FormatEnum"]])

slots.enum = Slot(uri=FRICTIONLESS.enum, name="enum", curie=FRICTIONLESS.curie('enum'),
                   model_uri=FRICTIONLESS.enum, domain=None, range=Optional[Union[str, List[str]]])

slots.resource = Slot(uri=FRICTIONLESS.resource, name="resource", curie=FRICTIONLESS.curie('resource'),
                   model_uri=FRICTIONLESS.resource, domain=None, range=Optional[str])

slots.fields = Slot(uri=FRICTIONLESS.fields, name="fields", curie=FRICTIONLESS.curie('fields'),
                   model_uri=FRICTIONLESS.fields, domain=None, range=Optional[Union[Union[str, FieldName], List[Union[str, FieldName]]]])

slots.reference = Slot(uri=FRICTIONLESS.reference, name="reference", curie=FRICTIONLESS.curie('reference'),
                   model_uri=FRICTIONLESS.reference, domain=None, range=Optional[Union[dict, Reference]])

slots.constraint_name = Slot(uri=FRICTIONLESS.constraint_name, name="constraint_name", curie=FRICTIONLESS.curie('constraint_name'),
                   model_uri=FRICTIONLESS.constraint_name, domain=None, range=Optional[str])

slots.missingValues = Slot(uri=FRICTIONLESS.missingValues, name="missingValues", curie=FRICTIONLESS.curie('missingValues'),
                   model_uri=FRICTIONLESS.missingValues, domain=None, range=Optional[Union[str, List[str]]])

slots.primaryKey = Slot(uri=FRICTIONLESS.primaryKey, name="primaryKey", curie=FRICTIONLESS.curie('primaryKey'),
                   model_uri=FRICTIONLESS.primaryKey, domain=None, range=Optional[Union[Union[str, FieldName], List[Union[str, FieldName]]]])

slots.foreignKeys = Slot(uri=FRICTIONLESS.foreignKeys, name="foreignKeys", curie=FRICTIONLESS.curie('foreignKeys'),
                   model_uri=FRICTIONLESS.foreignKeys, domain=None, range=Optional[Union[Union[dict, ForeignKeys], List[Union[dict, ForeignKeys]]]])

slots.profile = Slot(uri=FRICTIONLESS.profile, name="profile", curie=FRICTIONLESS.curie('profile'),
                   model_uri=FRICTIONLESS.profile, domain=None, range=Optional[Union[str, "ProfileEnum"]])

slots.title = Slot(uri=LINKML.title, name="title", curie=LINKML.curie('title'),
                   model_uri=FRICTIONLESS.title, domain=None, range=Optional[str])

slots.path = Slot(uri=FRICTIONLESS.path, name="path", curie=FRICTIONLESS.curie('path'),
                   model_uri=FRICTIONLESS.path, domain=None, range=Optional[str])

slots.dialect = Slot(uri=FRICTIONLESS.dialect, name="dialect", curie=FRICTIONLESS.curie('dialect'),
                   model_uri=FRICTIONLESS.dialect, domain=None, range=Optional[str])

slots.schema = Slot(uri=FRICTIONLESS.schema, name="schema", curie=FRICTIONLESS.curie('schema'),
                   model_uri=FRICTIONLESS.schema, domain=None, range=Optional[Union[dict, Schema]])

slots.resources = Slot(uri=FRICTIONLESS.resources, name="resources", curie=FRICTIONLESS.curie('resources'),
                   model_uri=FRICTIONLESS.resources, domain=None, range=Optional[Union[Dict[Union[str, ResourceName], Union[dict, Resource]], List[Union[dict, Resource]]]])

slots.range = Slot(uri=FRICTIONLESS.range, name="range", curie=FRICTIONLESS.curie('range'),
                   model_uri=FRICTIONLESS.range, domain=None, range=Optional[str])

slots.Reference_fields = Slot(uri=FRICTIONLESS.fields, name="Reference_fields", curie=FRICTIONLESS.curie('fields'),
                   model_uri=FRICTIONLESS.Reference_fields, domain=Reference, range=Optional[Union[Union[str, FieldName], List[Union[str, FieldName]]]])

slots.Reference_range = Slot(uri=FRICTIONLESS.range, name="Reference_range", curie=FRICTIONLESS.curie('range'),
                   model_uri=FRICTIONLESS.Reference_range, domain=Reference, range=Optional[str])

slots.ForeignKeys_fields = Slot(uri=FRICTIONLESS.fields, name="ForeignKeys_fields", curie=FRICTIONLESS.curie('fields'),
                   model_uri=FRICTIONLESS.ForeignKeys_fields, domain=ForeignKeys, range=Optional[Union[Union[str, FieldName], List[Union[str, FieldName]]]])

slots.ForeignKeys_range = Slot(uri=FRICTIONLESS.range, name="ForeignKeys_range", curie=FRICTIONLESS.curie('range'),
                   model_uri=FRICTIONLESS.ForeignKeys_range, domain=ForeignKeys, range=Optional[str])

slots.Schema_fields = Slot(uri=FRICTIONLESS.fields, name="Schema_fields", curie=FRICTIONLESS.curie('fields'),
                   model_uri=FRICTIONLESS.Schema_fields, domain=Schema, range=Optional[Union[Dict[Union[str, FieldName], Union[dict, Field]], List[Union[dict, Field]]]])



import logging
from typing import Union, Dict, Tuple, List, Any, Optional

from dataclasses import dataclass

from linkml.utils.schema_builder import SchemaBuilder
from linkml_runtime.linkml_model.meta import SchemaDefinition, SlotDefinition, EnumDefinition, \
    PermissibleValue, UniqueKey, ClassDefinition
from linkml_runtime.loaders import json_loader
from linkml_runtime.utils.formatutils import camelcase

from schema_automator.importers.import_engine import ImportEngine
import schema_automator.metamodels.frictionless as fl


TYPE_MAPPING = {
    "string": "string",
    "datetime": "datetime",
    "boolean": "boolean",
    "integer": "integer",
    "number": "decimal",
}


def _desc(elt: Union[fl.Field, fl.Resource]) -> Optional[str]:
    if elt.description:
        return elt.description[0]
    else:
        return None


def _add_unique_keys(cls: ClassDefinition, name: str, slot_names: List[str]):
    uk = UniqueKey(name, unique_key_slots=slot_names)
    cls.unique_keys[name] = uk


@dataclass
class FrictionlessImportEngine(ImportEngine):
    """
    An ImportEngine that imports Frictionless data packages with schema information

    See:

     `Frictionless specs <https://specs.frictionlessdata.io/>`_
     `Patterns <https://specs.frictionlessdata.io/patterns/>`_

    """

    def convert(self, file: str, id: str=None, name: str=None, **kwargs) -> SchemaDefinition:
        """
        Converts one or more JSON files into a Schema

        :param files:
        :param kwargs:
        :return:
        """
        package: fl.Package = json_loader.load(file, target_class=fl.Package)
        sb = SchemaBuilder()
        schema = sb.schema
        if id:
            schema.id = id
        if not name:
            name = package.name
        if name:
            schema.name = name
        schema.description = package.title
        for resource in package.resources:
            sb.add_class(resource.name)
            cls = schema.classes[resource.name]
            cls.description = _desc(resource)
            cls.title = resource.title
            tbl = resource.schema
            for field in tbl.fields:
                slot = SlotDefinition(field.name, description=_desc(field))
                cls.attributes[slot.name] = slot
                constraints = field.constraints
                if constraints:
                    slot.required = constraints.required
                    slot.pattern = constraints.pattern
                    if constraints.unique is True:
                        _add_unique_keys(cls, f"{slot.name}_unique_key", [slot.name])
                if field.enum:
                    e = self.add_enum(sb, field)
                    slot.range = e.name
                elif field.type:
                    t = str(field.type)
                    if field.type == fl.TypeEnum(fl.TypeEnum.array):
                        slot.multivalued = True
                    else:
                        slot.range = TYPE_MAPPING[t]
            if tbl.primaryKey:
                pks = tbl.primaryKey
                if len(pks) > 1:
                    _add_unique_keys(cls, f"{cls.name}_primary_key", [pks])
                else:
                    cls.attributes[pks[0]].identifier = True
            if tbl.foreignKeys:
                for fk in tbl.foreignKeys:
                    fk_fields = fk.fields
                    if isinstance(fk_fields, list) and len(fk_fields) > 1:
                        logging.warning(f"Cannot handle compound FKs: {cls.name}.[{fk_fields}]")
                    else:
                        if isinstance(fk_fields, list):
                            fk_field = fk_fields[0]
                        else:
                            fk_field = fk_fields
                        if fk_field:
                            fk_slot = cls.attributes[fk_field]
                            fk_slot.range = fk.reference.resource
                            # assume fk.fields is the PK
        sb.add_defaults()
        if name:
            schema.default_prefix = name
        for c in schema.classes.values():
            c.from_schema = 'http://example.org/'
        return sb.schema

    def add_enum(self, sb: SchemaBuilder, field: fl.Field) -> EnumDefinition:
        name = camelcase(f"{field.name}_enum")
        e = EnumDefinition(name)
        for code in field.enum:
            pv = PermissibleValue(code)
            # TODO: this behavior may be specific to C2M2, make this configurable
            if ":" in code:
                toks = code.split(":")
                if len(toks) == 2:
                    [prefix, short] = toks
                    pv = PermissibleValue(short, meaning=code)
                    sb.add_prefix(prefix, f"{sb.schema.id}/{prefix}/", replace_if_present=True)
            e.permissible_values[pv.text] = pv
        if e.name is sb.schema:
            raise NotImplementedError(f"Cannot yet merge enums")
        sb.schema.enums[e.name] = e
        return e







import logging
from pathlib import Path

from typing import Union, Dict, Tuple, List, Any, Type

from linkml.utils.schema_builder import SchemaBuilder
from linkml_runtime.linkml_model import SchemaDefinition, SlotDefinition

from dataclasses import dataclass

from sqlalchemy import create_engine, MetaData
import sqlalchemy.sql.sqltypes as sqlt

from schema_automator.importers.import_engine import ImportEngine

TYPE_MAP = {
    sqlt.TEXT: 'string',
    sqlt.INTEGER: 'integer',
    sqlt.FLOAT: 'float',
}

def _map_type(typ: Type) -> str:
    for k, v in TYPE_MAP.items():
        if isinstance(typ, k):
            return v
    return 'string'


@dataclass
class SqlImportEngine(ImportEngine):
    """
    An ImportEngine that introspects an SQL schema to determine a corresponding LinkML schema
    """
    engine: Any = None   ## TODO: determine typing info for SQL Alchemy

    def convert(
            self,
            file: str,
            name: str = None,
            model_uri: str = None,
            identifier: str = None,
            **kwargs
    ) -> SchemaDefinition:
        """
        Converts a SQL database

        :param file: path to SQLite OR an SQL Alchemy locator
        :param name:
        :param model_uri:
        :param identifier:
        :param kwargs:
        :return:
        """
        if '://' not in file:
            path = Path(file).resolve()
            locator = f"sqlite:///{path}"
        else:
            locator = file
        logging.info(f"Connecting to {locator}")
        engine = create_engine(locator)
        self.engine = engine
        metadata_obj = MetaData()
        logging.info(f"Reflecting using {engine}")
        metadata_obj.reflect(bind=engine)
        sb = SchemaBuilder()
        schema = sb.schema
        for table in metadata_obj.sorted_tables:
            logging.info(f"Importing {table.name}")
            sb.add_class(table.name)
            cls = schema.classes[table.name]
            pks = [column for column in table.columns if column.primary_key]
            if len(pks) == 1:
                pk = pks.pop().name
            else:
                pk = None
            for column in table.columns:
                slot = SlotDefinition(column.name)
                cls.attributes[slot.name] = slot
                if pk and pk == column.name:
                    slot.identifier = True
                if column.foreign_keys:
                    for fk in column.foreign_keys:
                        [fk_table, fk_table_col] = str(fk.column).split('.')
                        slot.range = fk_table
                else:
                    slot.range = _map_type(column.type)
        return schema





import logging

from dataclasses import dataclass
from tempfile import NamedTemporaryFile
from typing import List

from linkml_runtime.linkml_model.meta import SchemaDefinition
from schemasheets.schemamaker import SchemaMaker

from schema_automator.importers.import_engine import ImportEngine
import pandas as pd


@dataclass
class TableImportEngine(ImportEngine):
    """
    An ImportEngine that imports tabular data via schemasheets

    """
    element_type: str = None
    parent: str = None
    columns: List[str] = None

    def convert(self, file: str) -> SchemaDefinition:
        """
        Converts one or more JSON files into a Schema

        :param files:
        :param kwargs:
        :return:
        """
        df = pd.read_csv(file, sep='\t')
        self.import_from_dataframe(df)

    def import_from_dataframe(self, df: pd.DataFrame):
        """
        Imports a dataframe into a schema

        :param df:
        :return:
        """
        tf = NamedTemporaryFile(delete=False)
        if not self.columns:
            raise ValueError("Must specify columns")
        logging.info(f"Using columns: {self.columns}")
        ix = 1
        line = pd.DataFrame(dict(zip(df.head(), self.columns)), index=[ix])
        df = pd.concat([df.iloc[:ix-1], line, df.iloc[ix-1:]]).reset_index(drop=True)
        if self.parent:
            df.insert(0,
                      column="parent",
                      value=[f">{self.element_type}"] + [self.parent] * (len(df) - 1))
        df.to_csv(tf.name, sep='\t', index=False)
        #print(open(tf.name, 'r').read())
        #element_map = dict(zip(df.head(), self.columns))
        sm = SchemaMaker()
        return sm.create_schema([tf.name])



from schema_automator.annotators.schema_annotator import SchemaAnnotator
from schema_automator.annotators.jsonld_annotator import JsonLdAnnotator
from schema_automator.annotators.llm_annotator import LLMAnnotator



import click
import logging
import yaml
from dataclasses import dataclass
from typing import List, Union, Iterator

from linkml_runtime.linkml_model import SchemaDefinition, Element, PermissibleValue, ClassDefinition, SlotDefinition
from linkml_runtime.utils.metamodelcore import Curie
from linkml_runtime.utils.schemaview import SchemaView, re, EnumDefinition
from oaklib import BasicOntologyInterface
from oaklib.datamodels.search import SearchConfiguration
from oaklib.datamodels.text_annotator import TextAnnotation
from oaklib.interfaces import SearchInterface
from oaklib.interfaces.text_annotator_interface import TextAnnotatorInterface

from schema_automator.utils.schemautils import minify_schema

camel_case_pattern = re.compile(r'(?<!^)(?=[A-Z])')

def uncamel(n: str):
    # TODO: replace with equiv from linkml-runtime
    return camel_case_pattern.sub(' ', n).lower().replace('_', ' ')


@dataclass
class SchemaAnnotator:
    """
    An engine for enhancing schemas by performing lookup and annotation operations
    using an ontology service.

    A SchemaAnnotator wraps an OAK ontology interface.
    See `OAK documentation <https://incatools.github.io/ontology-access-kit/>`_ for more details
    """
    ontology_implementation: BasicOntologyInterface = None
    mine_descriptions: bool = False
    allow_partial: bool = False
    curie_only: bool = True
    assign_element_uris: bool = False
    assign_enum_meanings: bool = False

    def annotate_element(self, elt: Union[PermissibleValue, Element]) -> None:
        """
        Annotates an element or a permissible value

        :param elt:
        :return:
        """
        if isinstance(elt, Element):
            texts = [elt.name]
        elif isinstance(elt, PermissibleValue):
            texts = [elt.text]
        else:
            raise ValueError(f"Unexpected type {type(elt)}")
        texts += elt.aliases
        if self.mine_descriptions and elt.description:
            texts.append(elt.description)
        for text in texts:
            logging.info(f"Annotating: {text}")
            for r in self.annotate_text(text):
                logging.debug(f'MATCH: {r}')
                if self.allow_partial or r.matches_whole_text:
                    xref = r.object_id
                    if self.curie_only and not Curie.is_curie(xref):
                        continue
                    logging.info(f'Mapping from "{text}" to {xref}')
                    if isinstance(elt, PermissibleValue):
                        if self.assign_enum_meanings:
                            if not elt.meaning:
                                elt.meaning = xref
                                continue
                    else:
                        if self.assign_element_uris:
                            if isinstance(elt, ClassDefinition):
                                if not elt.class_uri:
                                    elt.class_uri = xref
                                    continue
                            if isinstance(elt, SlotDefinition):
                                if not elt.slot_uri:
                                    elt.slot_uri = xref
                                    continue
                            if isinstance(elt, EnumDefinition):
                                if not elt.enum_uri:
                                    elt.enum_uri = xref
                                    continue
                    if xref not in elt.exact_mappings:
                        elt.exact_mappings.append(xref)

    def annotate_text(self, text: str) -> Iterator[TextAnnotation]:
        # this is a wrapper over OAK annotation and search;
        # it (1) expands CamelCase (2) abstracts over annotation vs search
        # TODO: fold this functionality back into OAK
        oi = self.ontology_implementation
        text_exp = uncamel(text)  # TODO: use main linkml_runtime method
        if isinstance(oi, TextAnnotatorInterface):
            logging.debug(f"Using TextAnnotatorInterface on {text_exp}")
            # TextAnnotation is available; use this by default
            for r in oi.annotate_text(text_exp):
                yield r
            if text_exp != text.lower():
                for r in oi.annotate_text(text_exp):
                    yield r
        elif isinstance(oi, SearchInterface):
            logging.debug(f"Using SearchInterface on {text_exp}")
            # use search as an alternative
            cfg = SearchConfiguration(is_complete=True)
            for r in oi.basic_search(text, config=cfg):
                yield TextAnnotation(object_id=r, matches_whole_text=True)
            if text_exp != text.lower():
                for r in oi.basic_search(text_exp, config=cfg):
                    yield TextAnnotation(object_id=r, matches_whole_text=True)
        else:
            raise NotImplementedError

    def annotate_schema(self, schema: Union[SchemaDefinition, str]) -> SchemaDefinition:
        """
        Annotate all elements of a schema, adding mappings.

        This requires that the OntologyInterface implements either BasicOntologyInterface or SearchInterface
        """
        sv = SchemaView(schema)
        oi = self.ontology_implementation
        for elt_name, elt in sv.all_elements().items():
            self.annotate_element(elt)
        for e in sv.all_enums().values():
            for pv in e.permissible_values.values():
                self.annotate_element(pv)
        return sv.schema

    def enrich(self, schema: Union[SchemaDefinition, str]) -> SchemaDefinition:
        """
        Enrich a schema by performing lookups on the external ontology/vocabulary endpoint,
        and copying over metadata

        Currently, the only metadata obtained is text definitions

        .. code-block:: python

        >>> from schema_automator.annotators.schema_annotator import SchemaAnnotator
        >>> from oaklib.selector import get_implementation_from_shorthand
        >>> oi = get_implementation_from_shorthand("sqlite:obo:so")
        >>> sa = SchemaAnnotator(ontology_implementation=oi)
        >>> schema = sa.enrich("tests/data/schema.yaml")

        :param schema:
        :return:
        """
        sv = SchemaView(schema)
        oi = self.ontology_implementation
        for elt_name, elt in sv.all_elements().items():
            logging.debug(f"Enriching {elt_name}")
            if isinstance(elt, EnumDefinition):
                curies = []
                for pv in elt.permissible_values.values():
                    if pv.meaning:
                        pv_curies = [pv.meaning]
                    else:
                        pv_curies = []
                    self._add_description_from_curies(pv, pv_curies)
            else:
                curies = [sv.get_uri(elt)]
            for rel, ms in sv.get_mappings(elt_name).items():
                curies += ms
            self._add_description_from_curies(elt, curies)
        return sv.schema

    def _add_description_from_curies(self, elt: Union[Element, PermissibleValue], curies: List[str]):
        oi = self.ontology_implementation
        logging.info(f"Looking up descriptions using {curies}")
        for x in curies:
            logging.info(f"Fetching description using: {curies}")
            if elt.description:
                break
            try:
                defn = oi.definition(x)
                if defn:
                    elt.description = defn
                else:
                    mm = oi.entity_metadata_map(x)
                    logging.debug(f"MM={mm}")
                    for p in ['rdfs:comment', 'skos:definition', 'dcterms:description']:
                        if p in mm:
                            elt.description = mm[p]
                            last
            except Exception:
                pass


@click.command()
@click.argument('schema')
@click.option('--input', '-i', help="OAK input ontology selector")
@click.option('--output', '-o', help="Path to saved yaml schema")
def annotate_schema(schema: str, input: str, output: str, **args):
    """
    Annotate all elements of a schema.

    DEPRECATED: use main schemauto CLI instead
    """
    logging.basicConfig(level=logging.INFO)
    annr = SchemaAnnotator()
    schema = annr.annotate_schema(schema)
    sd = minify_schema(schema)
    if output:
        with open(output, 'w') as stream:
            yaml.safe_dump(sd, stream, sort_keys=False)
    else:
        print(yaml.safe_dump(sd, sort_keys=False))

if __name__ == '__main__':
    annotate_schema()


import json
from dataclasses import dataclass
from pathlib import Path

import click
import yaml
import logging
from copy import copy
from typing import Any, Tuple, Dict, Union, List, Optional

from deprecation import deprecated
from linkml.utils.schema_builder import SchemaBuilder
from linkml_runtime.linkml_model import SchemaDefinition, Element, ClassDefinition, \
    SlotDefinition, EnumDefinition, \
    ClassDefinitionName, \
    SlotDefinitionName, Prefix
from linkml_runtime.linkml_model.meta import ReachabilityQuery, AnonymousEnumExpression
from linkml_runtime.utils.formatutils import underscore, camelcase

from schema_automator.importers.import_engine import ImportEngine
from schema_automator.utils.schemautils import minify_schema, write_schema

# TODO: move to core. https://github.com/linkml/linkml/issues/104
RESERVED = ['in', 'not', 'def']

@dataclass
class JsonSchemaImportEngine(ImportEngine):
    """
    A :ref:`ImportEngine` that imports a JSON-Schema representation to a LinkML Schema
    """
    use_attributes: bool = False

    def convert(self, input: str, name=None, format = 'json', **kwargs) -> SchemaDefinition:
        """
        Converts a JSON-Schema json file into a LinkML schema

        :param input:
        :param name:
        :param format:
        :param kwargs:
        :return:
        """
        if format == 'json':
            with open(input) as stream:
                obj = json.load(stream)
        elif format == 'yaml':
            with open(input) as stream:
                obj = yaml.safe_load(stream)
        else:
            raise Exception(f'Bad format: {format}')
        return self.loads(obj, name, input_path=input, **kwargs)

    def import_project(self, import_directory: str, export_directory: str, match_suffix=".json", name=None, **kwargs) -> str:
        path = Path(import_directory)
        export_path = Path(export_directory)
        imports_list = []
        importer_name = "main.yaml"
        if name is None:
            name = "imported"
        path_to_schema_map = {}
        class_name_to_module_map = {}
        for item in path.rglob("*"):
            if str(item).endswith(match_suffix):
                relpath = item.relative_to(path)
                module_name = str(relpath.with_suffix("").as_posix())
                module_name_safe = "-".join(relpath.with_suffix("").parts)
                logging.info(f"Converting {item} => {module_name_safe}")
                schema = self.convert(str(item), name=module_name_safe, **kwargs)
                output_path = export_path / relpath
                output_path = output_path.with_suffix(".yaml")
                output_path.parent.mkdir(parents=True, exist_ok=True)
                path_to_schema_map[output_path] = schema
                if str(output_path) == importer_name:
                    raise ValueError(f"Must choose new importer_name: {output_path}")
                imports_list.append(module_name)
                for cn in schema.classes:
                    if cn in class_name_to_module_map:
                        raise ValueError(f"Class name clash; {cn} is in {module_name} and {class_name_to_module_map[cn]}")
                    class_name_to_module_map[cn] = module_name
        # add imports based on ranges
        for this_module, schema in path_to_schema_map.items():
            relpath = this_module.relative_to(export_path)
            logging.debug(f"Relative path {relpath} for module at: {this_module}")
            depth = len(relpath.parts) - 1
            rel = "../" * depth
            for cls in schema.classes.values():
                for a in cls.attributes.values():
                    rng = a.range
                    if rng not in class_name_to_module_map:
                        continue
                    if rng in schema.classes:
                        # no need to self-import
                        continue
                    import_module_name = class_name_to_module_map[rng]
                    if import_module_name not in schema.imports:
                        logging.info(f"Adding import to {import_module_name} in {schema.name} for {rng}")
                        schema.imports.append(import_module_name)
        for output_path, schema in path_to_schema_map.items():
            write_schema(schema, output_path)
        sb = SchemaBuilder(name=name)
        sb.add_defaults()
        s = sb.schema
        for i in imports_list:
            s.imports.append(i)
        importer_path = export_path / importer_name
        write_schema(s, importer_path)
        return str(importer_path)


    def load(self, input: str, name=None, format = 'json', **kwargs):
        return self.convert(input, name=name, format=format, **kwargs)

    def loads(self, obj: Any, name=None, **kwargs) -> SchemaDefinition:
        return self.translate_schema(obj, name, **kwargs)

    def _class_name(self, cn: str) -> str:
        # in future this can be configurable
        return camelcase(cn)

    def get_id(self, obj) -> str:
        if 'id' in obj:
            id = obj['id']
        elif '$ref' in obj:
            id = obj['$ref']
        else:
            raise Exception(f'No id {obj}')
        if id.startswith("#/definitions/"):
            return self.split_name(id)[0]
        else:
            return self._class_name(Path(str(id)).stem)

    def split_name(self, name) -> Tuple[str, str]:
        name = name.replace('#/definitions/', '')
        parts = name.split('.')
        name = parts.pop()
        pkg = '.'.join(parts)
        if pkg == '':
            pkg = None
        return name, pkg


    def translate_schema(self, obj: Dict, id_val=None, name=None, root_class_name=None, input_path: str = None) -> SchemaDefinition:
        if id_val is None and '$id' in obj:
            id_val = obj['$id']
        if id_val is None and '$schema' in obj:
            id_val = obj['$schema']
        if name is None and 'title' in obj:
            name = obj['title'].replace(' ','-')
            if name[0].isnumeric():
                name = f"_{name}"
        jsonschema_version = obj.get('$schema', None)
        if id_val is None and name is None:
            raise Exception(f'Must pass name OR id, or these must be present in the jsonschema')
        if name is None:
            name = id_val
        if id_val is None or not id_val.startswith('http'):
            id_val = f'https://example.org/{name}'
        name = underscore(name)
        self.schema = SchemaDefinition(id=id_val, name=name)
        if '$defs' in obj:
            self.translate_definitions(obj.get('$defs', {}))
        else:
            self.translate_definitions(obj.get('definitions', {}))
        if root_class_name is None:
            root_class_name = obj.get('name', None)
        if root_class_name is None:
            root_class_name = obj.get('title', None)
        if root_class_name is None and input_path:
            root_class_name = Path(input_path).stem
        if root_class_name is None:
            raise ValueError(f'No root class name: {obj}')
        root_class_name = self._class_name(root_class_name)
        if 'properties' in obj:
            logging.info(f'Root class: {root_class_name}')
            root_class = ClassDefinition(root_class_name)
            self.translate_properties(obj, root_class)
            self.schema.classes[root_class_name] = root_class
        self.schema.default_prefix = name
        self.schema.prefixes[name] = Prefix(name, id_val)
        self.schema.prefixes['linkml'] = Prefix('linkml', 'https://w3id.org/linkml/')
        self.schema.imports.append('linkml:types')
        return self.schema

    def translate_definitions(self, obj: Dict):
        for k, v in obj.items():
            self.translate_object(v, k)

    def translate_array(self, obj: Dict, name: str) -> SlotDefinition:
        if 'items' in obj:
            items_obj = obj['items']
            uniqueItems = obj.get('uniqueItems', False)
            if isinstance(items_obj, str):
                # found in DOSDP: TODO: check
                items_obj = {'type': 'string'}
            if 'properties' in items_obj:
                slot = SlotDefinition(name)
                slot.range = self.translate_object(items_obj)
            else:
                slot = self.translate_property(items_obj, name)
            if 'description' in items_obj:
                slot.description = items_obj['description']
            slot.multivalued = True
            return slot
        elif 'properties' in obj:
            c = ClassDefinition(f'{self._class_name(name)}Class')
            self.translate_properties(obj, c)
            self.schema.classes[c.name] = c
            slot = SlotDefinition(name)
            slot.range = c.name
            slot.multivalued = True
            return slot
        else:
            logging.error(f'NOT HANDLED: {obj} in array context')
            return None

    def translate_ref(self, obj: dict) -> ClassDefinitionName:
        return ClassDefinitionName(self._class_name(self.get_id(obj)))

    def translate_oneOf(self, oneOfList: List) -> Optional[ClassDefinition]:
        if all('$ref' in x for x in oneOfList):
            cns = [self.translate_ref(x) for x in oneOfList]
            n = self._class_name('_'.join(cns))
            c = ClassDefinition(n, union_of=cns)
            return c
        else:
            logging.warning(f'Cannot yet handle oneOfs without refs: {oneOfList}')
            return None

    def translate_property(self, obj: Dict, name: str, class_name: str = None) -> SlotDefinition:
        if name is None:
            raise ValueError(f'Name not set for {obj}')
        schema = self.schema
        aliases = []
        if name in RESERVED:
            aliases.append(name)
            name = f'_{name}'
        s = SlotDefinition(name)
        t = obj.get('type', None)
        s.description = obj.get('description', None)
        if s.description is not None:
            s.description = s.description.strip()
        # HCA-specific
        s.title = obj.get("user_friendly", None)
        default = obj.get('default', None)
        if '$ref' in obj:
            s.range = self.translate_ref(obj)
        elif t == 'array':
            s = self.translate_array(obj, name)
            if s is None:
                raise ValueError(f'Cannot translate array {name} {obj}')
        elif t == 'number':
            s.range = 'float'
        elif t == 'boolean':
            s.range = 'boolean'
        elif t == 'float':
            s.range = 'float'
        elif t == 'integer':
            s.range = 'integer'
            s.minimum_value = obj.get('minimum_value', None)
            s.maximum_value = obj.get('maximum_value', None)
        elif t == 'string':
            if 'enum' in obj:
                pvs = obj['enum']
                if self.use_attributes and class_name:
                    ename = f'{class_name}_{name}_options'
                else:
                    ename = f'{name}_options'
                schema.enums[ename] = EnumDefinition(name=ename, permissible_values=pvs)
                s.range = ename
            self._enum_from_ontology_extension(s, obj, name, class_name=class_name)
        else:
            logging.error(f'Cannot translate type {t} in {obj}')
        if s.name is schema.slots:
            logging.warning(f'TODO: unify alternate slots')
        if not self.use_attributes:
            schema.slots[s.name] = s
        return s

    def _enum_from_ontology_extension(self, slot: SlotDefinition, js_obj: dict, name: str, class_name: str = None):
        gr = js_obj.get("graph_restriction", None)
        if not gr:
            return
        if self.use_attributes and class_name:
            ename = f'{class_name}_{name}_options'
        else:
            ename = f'{name}_options'
        rqs = []
        for ont in gr["ontologies"]:
            rq = ReachabilityQuery(source_ontology=ont,
                                   source_nodes=gr.get("classes", []),
                                   include_self=gr.get("include_self", False),
                                   is_direct=gr.get("direct", False),
                                   relationship_types=gr.get("relations", []),
                                   )
            rqs.append(rq)

        if len(rqs) == 0:
            logging.warning(f"No ontologies in {gr}")
            return
        elif len(rqs) == 1:
            edef = EnumDefinition(ename, reachable_from = rqs[0])
        else:
            includes = [AnonymousEnumExpression(reachable_from=rq) for rq in rqs]
            edef = EnumDefinition(ename, include=includes)
        self.schema.enums[ename] = edef
        slot.range = ename

    def translate_object(self, obj: Dict, name: str = None) -> ClassDefinitionName:
        """
        Translates jsonschema obj of type object

        Generates a ClassDefinition, inserts it into the schema, and returns the name
        """
        schema = self.schema
        t = obj.get('type', None)
        desc = obj.get('description', None)
        additional = obj.get('additionalProperties', False) ## TODO
        if 'oneOf' in obj:
            unionCls = self.translate_oneOf(obj.get('oneOf'))
        else:
            unionCls = None
        allOf = obj.get('allOf', None)
        title = obj.get('title', None)
        required = obj.get('required', [])

        properties = obj.get('properties', {})
        discriminator = obj.get('discriminator', None)  # In OpenAPI but not JSON-Schema
        pkg = None
        if name is None:
            name = obj.get('title', None)
        else:
            name, pkg = self.split_name(name)
            if name is None:
                raise ValueError(f'Problem splitting name from package')
        if name is None:
            name = 'TODO'
        name = self._class_name(name)
        c = ClassDefinition(name, description=desc, from_schema=pkg)
        if unionCls:
            c.union_of = unionCls.union_of
        for k, v in properties.items():
            # TODO: reuse below
            slot = self.translate_property(v, k, class_name=name)
            if slot.name in required:
                slot.required = True
            if self.use_attributes:
                c.attributes[slot.name] = slot
            else:
                c.slots.append(slot.name)
                c.slot_usage[slot.name] = slot
        schema.classes[c.name] = c
        return c.name


    def translate_properties(self, obj: dict, parent_class: ClassDefinition):
        required = obj.get('required', [])
        for k, v in obj.get('properties',{}).items():
            slot = self.translate_property(v, k, class_name=parent_class.name)
            if slot.name in required:
                slot.required = True
            if self.use_attributes:
                parent_class.attributes[slot.name] = slot
            else:
                parent_class.slots.append(slot.name)
                parent_class.slot_usage[slot.name] = slot


@click.command()
@click.argument('input')
@click.option('--name', '-n', required=True, help='ID of schema')
@click.option('--format', '-f', default='json', help='JSON Schema format - yaml or json')
@click.option('--output', '-o', help='output path')
def jsonschema2model(input, output, name, format, **args):
    """ Infer a model from JSON Schema """
    ie = JsonSchemaImportEngine()
    schema = ie.load(input, name=name, format=format)
    write_schema(schema, output)


if __name__ == '__main__':
    jsonschema2model()




